/*
    Maciej Gugaa, H1TIZ
    
    Program: przykad uycia kontenera queue.
    
    Opis: 
          Szablon queue to jesten z adaptatorw kontenerw sekwencyjnych, czyli takich ktre przystosowuj
          standardowe kontenery STL do specjalnych celw. Szablon klasy pozwala danej klasie (domylnie deque)
          udostpnia typowy interfejs kolejki typu FIFO.
          
          Szablon queue w odrnieniu do deque nie pozwala na dostp swobodny do elementw kolejki, ani na iteracje
          po kontenerze. Moliwe s podstawowe operacje: dodawanie elementu na koniec kolejki, usuwanie elementu z 
          pocztku kolejki, sprawdzenie wartoci pocztkowego i kocowego elementu, okrelenie liczby elementw oraz
          sprawdzenie czy kolejka jest pusta.
          
          Interfejs kolejki z uwzgldnieniem wewntrznego interfejsu kolejki:
                    
                    +---------------------------------------------------------------------+																  |
                    |			         back()  /\			/\  front()		  		      |
                    |					         ||			||				  		  	  |
                    |					         \/			\/				  			  |
                    |           +---------------------------------------------+   	      |
                    |  push()   |				  ## queue ##				  |  pop()    |
                    |			|											  |			  |
                    |  =======> |			   back()	  front()		  	  |	=======>  |
                    |			|  push_back()   /\         /\    pop_front() |           |
                    |			|	   /\		 ||			||		  /\	  |			  |
                    |			|	   ||		 ||			||		  ||	  |			  |
                    |			|	   \/		 \/         \/        \/      |			  |
                    |			|	+------------------------------------+	  | 		  |
                    |			|   |  ## kontener(domylnie: deque) ##	 |	  |			  |
                    |			|	+------------------------------------+	  |			  |
                    |			+---------------------------------------------+			  |
                    +---------------------------------------------------------------------+
    
    Przykad uycia:
          Pewna firma chce otworzy sklep. Szef firmy zastanawia si, czy istnieje potrzeba stworzenia dwch
          stanowisk sprzedarzy towarw, czy wystarczy tylko jedno. Naszym zadaniem jest napsanie prostego 
          programy obrazujcego jak bdzie przebiegaa ogsuga klientw w przypadku gdy posiadamy jetno stanowisko
          sprzeday: ilu klientw zostanie obsuonych, jak dugi bdzie czas oczekiwania na obsug, ile osb 
          zrezygnuje z dokonania zakupw w tym sklepie w zwizku ze zbyt dug kolejk. Dla uproszczenia zakadamy e:
           - klient rezygnuje z oczekiwania na dokonanie zakupu gdy w kolejce stoi wiecej ni 10 osb
           - obsuga klienta trwa od 2 do 10 minut, przy czym 1/8 klientw jest obsuona w cigu 2 minut, 1/8 w cigu 3, itd.
           - odstpy czasowe pomidzy pojawianiem si nowych klientw w kolejce s losowe, ale rednia na godzin wychodzi staa
*/

/* Doczamy plik nagwkowy <queue> aby korzysta z kolejki */
#include <queue>
/* Inteligentny wskanik scoped_ptr z biblioteki boost */
#include <boost/shared_ptr.hpp>

#include <iostream>
#include <string>

using namespace std;
/* Przykadowa klasa reprezentujca klienta */
class Client
{
      private:
              long   arrive;       // Czas doczenia klienta do kolejki
              int    serviceTime;  // Czas obsugi klienta (losowa liczba cakowita z zakresu 2-5)
      public:
                     Client(long arr) : arrive(arr) {serviceTime = std::rand() % 3 + 2;}
                     ~Client() {}
              long   getArrive() const {return arrive;}
              int    getServiceTime() const {return serviceTime;}
};
/* Zwraca true gdy nowy klient stan w kolejce */
bool newClient(double x)
{
    return (std::rand() * x / RAND_MAX < 1);
}
/*
  clients jest kolejk inteligentnych wskanikw typu Client. Zamiast uywa inteligentnych wskanikw moglibymy
  zdefiniowa typ:
              
              typedef queue<Client *> clients;
  
  Ma to jednak wad, gdy podczas usuwania elementw z kolejki obiekty nie byyby automatycznie niszczone - sami 
  musielibymy pamita, eby je zniszczy. Problem taki nie wystpiby gdybymy uyli typu:
             
             typedef queue<Client> clients;
  
  Wowczas jednak gdybymy w przyszoci zdefiniowali inne typy klientw dziedziczce po klasie Client, nie moglibymy
  ich umieszcza w tej kolejce. Dlatego opaca si uywa inteligentnych wskanikw. Naley jednak uwaa, gdy 
  nie wszystkie z nich nadaj si do tego celu (np. std::auto_ptr(transfer wasnoci przy kopiowaniu), boost::scoped_ptr
  (brak moliwoci kopiowania))
*/
typedef boost::shared_ptr<Client> client_ptr;
typedef queue<client_ptr> clients;

int main(void)
{   
    srand(time(0)); // generator liczb losowych
    clients shopClients;
    
    cout << "Badanie czasu obslugi klientow." << endl 
         << "Podaj srednia liczbe klientow na godzine: ";
    double clientsPerHour;
    cin >> clientsPerHour;
    
    /* Ustawienie parametrw symulacji */
    /* Symulacja z jednominutow rozdzielczoci, zakadamy e sklep jest otwarty przez 10 godzin dziennie */
    long   cycles = 60 * 10;                                // liczba cykli
    double minPerClient = 60 / clientsPerHour;              // redni odstp czasowy midzy klientami
    long   turnaways = 0;                                   // liczba klientw nieobsuonych
    long   allClients = 0;                                  // liczba klientw przyjtych do kolejki
    long   served = 0;                                      // liczba klientw obsuonych
    long   allWaitingClients = 0;                           // czna liczba klientw oczekujcych
    int    waitTime = 0;                                    // czas do koca obsugi danego klienta
    long   allTime = 0;                                     // czny czas oczekiwania
    
    // symulacja
    for (int cycle = 0; cycle < cycles; cycle++)
    {
        if (newClient(minPerClient)) // gdy pojawi si nowy klient
        {
            if (shopClients.size() > 10) // size() zwraca liczb elementw w kolejce
                turnaways++;             // klient rezygnuje, musiaby za dugo czeka
            else
            {
                /* 
                   Dodanie nowego klienta na koniec kolejki,
                   metoda push(const T& x) wstawia element na koniec kolejki 
                */
                shopClients.push(client_ptr(new Client(cycle)));
                allClients++;
            }
        }
        /*
          Metoda empty() zwraca true gdy kolejka jest pusta, w przeciwnym razie zwraca false
          Przed wywoaniem funkcji skadowej nalezy upewnic sie, ze kolejka zawiera jakiekolwiek
          elementy. W innym przypadku jej dziaanie jest nieokreslone.
        */
        if (waitTime <= 0 && !shopClients.empty())
        {
            /*
              Metoda T& front() zwraca referencj do elementu na pocztku kolejki
            */
            client_ptr nowService = shopClients.front();         // aktualnie obsugiwany klient
            waitTime = nowService->getServiceTime();             // czas obsugi klienta
            allTime += cycle - nowService->getArrive();          // czas oczekiwania w kolejce
            served++;
            /*
              Metoda pop() usuwa element z kolejki
            */
            shopClients.pop();
            /*
              Gdybymy nie uyli uyli inteligentnych wskanikw, tylko:
                  
                  typedef queue<Client *> clients;
        
              musielibymy sami pamita o zwolnieniu pamici po usunitym z kolejki elemencie:
                  
                  Client * nowService = clients.front();
                  ... (przetwarzanie nowService)
                  clients.pop();
                  delete nowService;
            */
        }
        if (waitTime > 0)
           waitTime--;
        allWaitingClients += shopClients.size();
    }
    
    if (allClients > 0)
    {
        cout << "Liczba przyjetych klientow: " << allClients << endl
             << "Liczba klientow obsluzonych: " << served << endl
             << "Liczba klientow odeslanych: " << turnaways << endl
             << "Srednia dlugosc kolejki: ";
        cout.precision(2);
        cout.setf(ios_base::fixed, ios_base::floatfield);
        cout.setf(ios_base::showpoint);
        cout << (double) allWaitingClients / cycles << endl
             << "Sredni czas oczekiwania: "
             << (double) allTime / served << " minut\n";
    }
    else
        cout << "Brak klientow\n";
        
    cout << "Koniec symulacji.\n";
    /*
        Jeeli si zdarzy, e jacy klienci stali w kolejce, ale upyno 10 godzin i sprzedawca stwierdzi
        e zamyka sklep, obiekty reprezentujce klientw zostan automatycznie zniszczone. Gdybymy uyli:
                  
                  typedef queue<Client *> clients;
        
        zamiast inteligentnych wskanikw, musielibymy je sami zwolni pami po usuwanych elementach:
                  
                  while (clients.size())
                  {
                      Client * temp = clients.front();
                      clients.pop();
                      delete temp;
                  }
    */
    return 0;
}
